import asyncio
from functools import partial
from typing import cast
from unittest import TestCase
from unittest.mock import Mock, call, sentinel

from testfixtures import ShouldWarn

from frgal.aio import AsyncGrpcClient
from frgal.client import GrpcClient
from frgal.decoder import GrpcDecoder
from frgal.utils import (MESSAGE_MAX_LENGTH, AsyncMock, AsyncTestClientMixin, CallableMock, MockingChannel,
                         TestClientMixin, format_message, get_call_id, make_awaitable)

from .proto.example_pb2 import Reply, SimpleMessage

try:
    from unittest import IsolatedAsyncioTestCase  # type: ignore
except ImportError:  # pragma: no cover
    from asynctest import TestCase as IsolatedAsyncioTestCase  # type: ignore


class GetCallIdTest(TestCase):
    def test_get_call_id(self):
        call_id = get_call_id()

        self.assertIsInstance(call_id, str)
        self.assertEqual(len(call_id), 4)

    def test_random(self):
        self.assertNotEqual(get_call_id(), get_call_id())


class FormatMessageTest(TestCase):
    def test_short_message(self):
        message = SimpleMessage()
        message.value = "I am fish!"
        self.assertEqual(format_message(message), 'value: "I am fish!"')

    def test_long_message(self):
        message = SimpleMessage()
        message.value = "I am fish! " * 400

        out = format_message(message)

        self.assertEqual(len(out), MESSAGE_MAX_LENGTH)
        self.assertRegex(out, r'^value: "I am fish! .*\.\.\.$')

    def test_short_object(self):
        self.assertEqual(format_message(42), '42')

    def test_long_object(self):
        message = ["I am fish!"] * 400
        self.assertRegex(format_message(message), r"^\['I am fish!', .*\.\.\.$")

    def test_generator(self):
        obj = ("I am fish!" for i in range(400))  # pragma: no cover
        self.assertRegex(format_message(obj), r'^<generator object .*>$')


class CallableMockTest(TestCase):
    def test_call(self):
        a_callable = Mock(return_value=sentinel.reply)
        a_call = CallableMock('/Coffee/make', a_callable)

        self.assertEqual(a_call(sentinel.request), sentinel.reply)

        self.assertEqual(a_callable.mock_calls, [call(sentinel.request)])


class MockingChannelTest(TestCase):
    def _test_callable(self, name: str) -> None:
        client = Mock(name='client')
        channel = MockingChannel(client)

        a_callable = getattr(channel, name)('/Coffee/make')

        a_call = cast(partial, a_callable.call)
        self.assertEqual(a_call.func, client.mock)
        self.assertEqual(a_call.args, ())
        self.assertEqual(a_call.keywords, {'method': '/Coffee/make'})

    def test_unary_unary(self):
        self._test_callable(name='unary_unary')

    def test_unary_stream(self):
        self._test_callable(name='unary_stream')

    def test_stream_unary(self):
        self._test_callable(name='stream_unary')

    def test_stream_stream(self):
        self._test_callable(name='stream_stream')


class CoffeeStub:
    """A simple stub for testing."""

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.channel = channel
        self.make = channel.unary_unary('/Coffee/make')


class TestClient(TestClientMixin, GrpcClient):
    """Simple test client to test a TestClientMixin."""


class TestClientMixinTest(TestCase):
    def test_connect(self):
        client = TestClient(sentinel.netloc, sentinel.decoder, [])

        client.connect()

        self.assertIsInstance(client.channel, MockingChannel)

    def test_call(self):
        client = TestClient(sentinel.netloc, GrpcDecoder(), [Mock(CoffeeStub=CoffeeStub)])
        request = SimpleMessage(value='10')
        reply = Reply()
        reply.data.result = 'Espresso!'
        client.mock.return_value = reply

        result = client.call('Coffee', 'make', request)

        self.assertEqual(result, 'Espresso!')
        self.assertEqual(client.mock.mock_calls, [call(request, method='/Coffee/make', timeout=None)])

    def test_call_future(self):
        client = TestClient(sentinel.netloc, GrpcDecoder(), [Mock(CoffeeStub=CoffeeStub)])
        request = SimpleMessage(value='10')
        reply = Reply()
        reply.data.result = 'Espresso!'
        client.mock.return_value = reply

        result = client.call_future('Coffee', 'make', request)

        self.assertEqual(result.result(), 'Espresso!')
        self.assertEqual(client.mock.mock_calls, [call(request, method='/Coffee/make', timeout=None)])

    def test_call_stream(self):
        client = TestClient(sentinel.netloc, GrpcDecoder(), [Mock(CoffeeStub=CoffeeStub)])
        request = SimpleMessage(value='10')
        reply = Reply()
        reply.data.result = 'Espresso!'
        client.mock.return_value = [reply, reply]

        result = client.call_stream('Coffee', 'make', request)

        self.assertEqual(tuple(result), ('Espresso!', 'Espresso!'))
        self.assertEqual(client.mock.mock_calls, [call(request, method='/Coffee/make', timeout=None)])


class AsyncTestClient(AsyncTestClientMixin, AsyncGrpcClient):
    """Simple test client to test a AsyncTestClientMixin."""


class AsyncTestClientMixinTest(IsolatedAsyncioTestCase):
    def test_init(self):
        client = AsyncTestClient(sentinel.netloc, GrpcDecoder(), [Mock(CoffeeStub=CoffeeStub)])
        self.assertIsInstance(client.mock, AsyncMock)
        self.assertIsInstance(client.channel, MockingChannel)
        self.assertIs(client.channel._loop, asyncio.get_event_loop())


class AsyncMockTest(IsolatedAsyncioTestCase):
    def setUp(self):
        self.mock = AsyncMock(spec=('call',))

    async def test_awaitable(self):
        self.mock.call.return_value = make_awaitable(sentinel.result)
        self.assertEqual(await self.mock.call(), sentinel.result)

    async def test_async_iterable(self):
        async def async_gen():
            yield sentinel.result
        self.mock.call.return_value = async_gen()
        self.assertEqual([i async for i in self.mock.call()], [sentinel.result])

    async def test_deprecation(self):
        self.mock.call.return_value = sentinel.result
        with ShouldWarn(DeprecationWarning(
            'Automagical transformation to awaitables is deprecated '
            'and will be removed in the next major release.')
        ):
            self.assertEqual(await self.mock.call(), sentinel.result)
